/***************************************************************************
   dict.cpp  -  talks to the dict server dictd: sending and receiving data
                         -------------------
   Copyright (C) 2006-2008 by Helge Kortendieck alan@kasmithee.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "dict.h"

dict::dict( QObject *parent ) : QObject( parent )
{
  bDictLookUpWordBlocked = false;           // do not start another word lookup before running finished
  iConState = 0;                            // connection to dict server state: 0=disconnected, 1=connected, 2=trying to connect
  iRecon = 1;                               // init disconnecting, assume final disconnnect (not reconnect (=1)) as default
  sLookUp = "";

  sock = new QTcpSocket( this );
  connect( sock, SIGNAL( connected() ), this, SLOT( conConnected() ) );       // receive connected to dict server
  connect( sock, SIGNAL( disconnected() ), this, SLOT( conDisConnected() ) ); // receive disconnected from dict server
  connect( sock, SIGNAL( readyRead() ), this, SLOT( socketReadyRead() ) );

  timerCom = new QTimer(this);              // 3 sec timer to monitor whether a command to the dict server timed out - unblock then
  timerCom->setSingleShot( true );
  connect(timerCom, SIGNAL(timeout()), this, SLOT( timerComTimedOut() ) );

  timerCon = new QTimer(this);              // 2 sec timer to catch if a connection to dict server times out
  timerCon->setSingleShot( true );
  connect(timerCon, SIGNAL(timeout()), this, SLOT( conConnectingTimedOut() ) );

//  timerDisCon = new QTimer(this);           // 1 sec timer, to force disconnection after asking nicely timed out
//  timerDisCon->setSingleShot( true );
//  connect(timerDisCon, SIGNAL(timeout()), this, SLOT( conDisConForce() ) );

  init();
}

dict::~dict()
{
}


void dict::init()                           // read settings, try to connect to dict server
{
  iTask = 1;                                    // current task: 0=idle, 1=connect to database n send dict list to main, 2=word lookup, 3=send dict list to set dlg
  settingsRead();                            // 0=do not reconnect, disconnect before connect
  con();
}

void dict::settingsDlgUpd()                 // settings dialog was updated (saved), re-connect, re-read settings and check dicts
{
  iTask = 1;                                   // current task: 0=idle, 1=connect to database n send dict status to main, 2=word lookup, 3=send dict list to set dlg
  iRecon = 1;
  settingsRead();                            // 1=re-read after changes in settings dialog (disconnect first!)
  conDisCon( 1 );
}


// ****************************************************************************
// **  (re-)read dict server relevant settings from config file              **
// ****************************************************************************
void dict::settingsRead()
{
  dictsInConf.clear();
  QSettings settings("qvtwfc", "qvtwfc");

  settings.beginGroup( "dict" );
    sDictIp = settings.value( "dictip", "localhost").toString();
    iDictPort = settings.value( "dictport", 2628 ).toInt();
    iD = 0;
    iD = settings.beginReadArray("dicts");
    if( iD > 0 ) {
      for (int i = 0; i < iD; i++) {
        settings.setArrayIndex(i);
        dictsEntry.selected = true;
        dictsEntry.name = settings.value( "dictName" ).toString();
        dictsEntry.desc = settings.value( "dictDesc" ).toString();
        dictsInConf.append( dictsEntry );
      }
    }
    settings.endArray();
  settings.endGroup();
}


// ****************************************************************************
// **  connect/disconnect to/from dict server                                **
// ****************************************************************************
void dict::con()                            // connect to dict server
{
//  cout << "con(): reached ! " << endl;                                     // *****  debug  *****
  iConState = 2;                                // connection to dict server state: 0=disconnected, 1=connected, 2=trying to connect
  if( sock->state() != QAbstractSocket::UnconnectedState ) {  // if we are still connected in any kind, close current connection
    sock->abort();
  }
  timerCon->start(2000);
  sock->connectToHost( sDictIp, iDictPort );
}

void dict::conConnected()                   // receive connected to dict server
{
//  cout << "conConnected(): reached !" << endl;                                     // *****  debug  *****
  if( timerCon->isActive() == true ) {
    timerCon->stop();
  }
  iConState = 1;
  if( iTask == 1 or iTask == 3 ) {
    getDictsFromServer();
  } else if( iTask == 2 and sLookUp.length() > 0 ) {
    lookUpWordLoopStart();
  }
}

void dict::conConnectingTimedOut()          // connecting to dict server timed out
{
//  cout << "conConnectingTimedOut(): reached ! " << endl;                                     // *****  debug  *****
  if( sock->state() != QAbstractSocket::ConnectedState ) {
    iConState = 0;
    sock->abort();
    sErr = tr( "Timed out while connecting to dict server dictd - dictionaries unavailable !" );
    emit sigDicErrCon( iTask, sErr );
    iTask = 0;                                  // current task: 0=idle, 1(2)3 = connect and: 1=check dicts avail, 2=word lookup, 3=send dict list to set dialog
    iRecon = 0;                                 // do not go into an eternal loop
    bDictLookUpWordBlocked = false;
  }
}

void dict::conReCon( const QString& sD, int iD) // re-connect to dict server with new settings
{
  sDictIp = sD;
  iDictPort = iD;
  iTask = 3;
  conDisCon( 1 );
//  cout << "conReCon(): reached ! " << endl;                                     // *****  debug  *****
}

void dict::conDisCon( int iD )              // disconnect from dict server, iRecon=0 disconnect , =1 reconnect
{
//  cout << "conDisCon(): reached ! " << endl;                                     // *****  debug  *****
  iRecon = iD;
  if( sock->state() == QAbstractSocket::ConnectedState ) {
    sock->disconnectFromHost();
//    connect(timerDisCon, SIGNAL(timeout()), this, SLOT( conDisConForce() ) );
//    timerDisCon->start(1000);                   // 1 sec window to allow to disconnect
  } else {
    conDisConnected();                          // ? socked died ?, needed ??
  }
}

void dict::conDisConnected()                // receive disconnected from dict server
{
//  cout << "conDisConnected(): reached !" << endl;                                     // *****  debug  *****
//  if( timerDisCon->isActive() == true ) {
//    disconnect(timerDisCon, SIGNAL(timeout()), this, SLOT( conDisConForce() ) );
//    timerDisCon->stop();                          // 1 sec window to allow to connect     !!!!!  will not do, always emits signal 'timeout'  !!!!!
//  }
  iConState = 0;
  bDictLookUpWordBlocked = false;
  if( iRecon == 1 ) {
    con();
  }
}

void dict::conDisConForce()                 // hard close the socket
{
//  cout << "conDisConForce2(): reached ! " << endl;                                     // *****  debug  *****
  if( sock->state() != QAbstractSocket::UnconnectedState ) {
    sock->abort();                              // should trigger 'disconnected' signal -> conDisConnected()
//    iConState = 0;
//    if( iRecon == 1 ) {
//      con();
//    }
  }
}


// ****************************************************************************
// **  timers                                                                **
// ****************************************************************************
void dict::timerComStart()
{
  if( timerCom->isActive() == false ) {
    timerCom->start(3000);
  }
}

void dict::timerComStop()
{
  if( timerCom->isActive() == true ) {
    timerCom->stop();
  }
}

void dict::timerComTimedOut()
{
  if( sock->state() != QAbstractSocket::ConnectedState ) {
    bConnected = false;
    sock->abort();
    bDictLookUpWordBlocked = false;
    sErr = tr( "Timed out while sending a command to the dict server (3 sec window) - seems we lost the dict server !\n" );
    emit sigDicErrQuery( iErr, sErr );
  }
}


// *******************************************************************************************************
// **  query dict server: get list of avail. dictionaries, get info abouta dictionary, get translation  **
// *******************************************************************************************************
void dict::getDictsFromServer()             // show available databases, 1=for settings dialog, 2=at start, keep quiet, do not send list
{
  iCom = 0;                                     // "SHOW DB"
  dictsOnServer.clear();
  sendCommand();
}

void dict::lookUpWordPre( const QString& sD )  // prepare look up word, need to re-connect?
{
//  cout << "lookUpWordPre(): reached ! " << sD.toStdString() << endl;                                     // *****  debug  *****
  sLookUp = sD;                                     // text to look up, received from qvtwfc
  if( bDictLookUpWordBlocked == false ) {
    lookUpWordLoopPre();
  }
}


void dict::lookUpWordLoopPre()              // prepare looping over the avialable dicts
{
//  cout << "lookUpWordLoopPre(): reached ! " << endl;                                     // *****  debug  *****
  if( bDictLookUpWordBlocked == false ) {
    if( sLookUp.length() > 0 ) {
      iTask = 2;
      if( sLookUp.length() == 1 ) {
        conDisCon( 1 );                           // re-connect to dict server, then look up word
      } else {
        lookUpWordLoopStart();
      }
    }
  } else {
    QTimer::singleShot(200, this, SLOT( lookUpWordLoopPre() ) );  // short timer to wait for lookup result before next request can be made
  }
}

void dict::lookUpWordLoopStart()            // look up word in next dictionary or send result
{
//  cout << "lookUpWordLoopStart(): reached ! " << endl;                                     // *****  debug  *****
  if( dictsToQuery.count() > 0 ) {          // don't start lookup if no dictionary available
    bDictLookUpWordBlocked = true;
    iDictToQuery = 0;
    lookups.clear();                        // clear QList with result struct
    sDb = dictsToQuery.at( 0 ).name;        // get 1st dictionary to query
    if( iConState == 1 ) {                  // only lookup word, if connected
      lookUpWord();                             // look up word, loop over avail dicts
    }
  }
}

void dict::lookUpWordLoopNext()             // look up word in next dictionary or send result
{
//  cout << "lookUpWordLoopNext(): reached ! " << endl;                                     // *****  debug  *****
  iDictToQuery = iDictToQuery + 1;
  if( iDictToQuery < dictsToQuery.count() ) {
    sDb = dictsToQuery.at( iDictToQuery ).name;  // get next dictionary to query
//    cout << "lookUpWordLoopNext(): lookUpWord reached ! " << sDb.toStdString() << " : " << sLookUp.toStdString() << endl;                // *****  debug  *****
    lookUpWord();
  } else {                                      // no more dictionaries to query send result
//    cout << "lookUpWordLoopNext(): loop end reached ! " << endl;                                     // *****  debug  *****
    bDictLookUpWordBlocked = false;
    emit sigDicSendMatches( lookups );
  }
}

void dict::lookUpWord()                     // look up word
{
//  cout << "lookUpWord(): reached ! " << sDb.toStdString() << " : " << sLookUp.toStdString() << endl;                 // *****  debug  *****
  ssRead = "";                                                          // ******************  debug  ********************
  iCom = 2;                                         // 2=DEFINE
  sendCommand();
}


// ****************************************************************************
// **  dict server: send command to / receive answer from                    **
// ****************************************************************************
void dict::sendCommand()                    // send command to dict server, 0=show databases, 1=show db xy info, 2=define <def.>, 3=match <definition>
{
  iResType = 0;                                 // response type: 0=undefined_screwed up, 1=databases, 2=database info, 3=word match orig, 4=word match trans
  iDictQueryErr = 0;
  timerComStart();                              // start command timeout timer (3 sec)
  switch ( iCom ) {
    case 0:                                     // show databases
      sSend = "SHOW DB";
      break;
    case 1:                                     // show info to one database
      sSend = "SHOW INFO";
      break;
    case 2:                                     // lookup via DEFINE command
      sSend = "DEFINE " + sDb + " " + sLookUp;
      break;
    case 3:                                     // lookup via MATCH command
      sSend = "MATCH " + sDb + " . " + sLookUp;     // '.' == use server default strategy
      break;
  }
  QTextStream os(sock);                         // write to the router
  os << sSend << "\n";
}

void dict::socketReadyRead()                // read server answer
{
  timerComStop();
  while ( sock->canReadLine() ) {
    sRead = QString::fromUtf8( sock->readLine() );
    sRead = sRead.trimmed();                    // deletes cr lf
    ssRead = ssRead +  "from Dictd: " + sRead + "\n";                                                                                      // *********** debug
    iD = sRead.left(3).toInt( &bConv, 10 );
    if( bConv  and sRead.left( 1 ) != " " ) {                           // got a response code
      iRCode = iD;                              // response code from dict server
 //       ssRead = ssRead + "*z*" + QString::number( iRCode );                                                                                      // *********** debu
      switch ( iRCode ) {
        case 110:                               // n databases present - text follows - answer to "SHOW DB"
          iResType = 1;                         // response type: 0=undefined_screwed up, 1=databases, 2=database info, 3=word match orig, 4=word match trans
        break;
        case 112:                               // database information follows - answer to client "SHOW INFO" command
          iResType = 2;                         // response type: 0=undefined_screwed up, 1=databases, 2=database info, 3=word matches, 4=word matches from MATCH
        break;
        case 150:                               // n definitions retrieved - definitions follow
          break;
        case 151:                               // word database name - text follows
          lookup.trans = "";
          iResType = 3;                         // response type: 0=undefined_screwed up, 1=databases, 2=database info, 3=word match orig, 4=word match trans
          iD = sRead.indexOf( " " );
          iD2 = sRead.indexOf( " ", iD + 1 );
          iD3 = sRead.indexOf( " ", iD2 + 1 );    // expecting '# result database database-info' line
          lookup.dict = sRead.mid( iD2 + 1, iD3 - iD2 - 1 );// extract database/dictionary name for this result
          ssRead = ssRead + "lookup.dict= " + sRead.mid( iD2 + 1, iD3 - iD2 - 1 ) + "\n";
          break;
        case 152:                               // n matches found - text follows
          ssRead = "";                          // ***************************  debug  **************************************************************************
          lookups.clear();
          iResType = 4;                         // response type: 0=undefined_screwed up, 1=databases, 2=database info, 3=word matches, 4=word matches from MATCH
          break;
        case 220:                               // text msg-id, send by dict server on successful connect
          sDictd = sRead;
          bConnected = true;
//          showDb();                             // connected, now get available databases...
          break;
        case 221:                               // Closing connection - answer to client "QUIT" command
          // iD = sRead.toInt();
          break;
        case 250:                               // ok - command complete
          switch ( iTask ) {                        // current task: 0=idle, 1=connect to database n send dict list to main, 2=word lookup, 3=dict list to set dialog
            case 1:                                 // 1=databases - answer to "SHOW DB", check dicts avail of selected in settings
              checkDictsAvail();
              break;
            case 2:                                 // 2=word matches from DEFINE
              lookUpWordLoopNext();
              break;
            case 3:                                 // 3=databases - answer to "SHOW DB", send list to settings dialog
//              sendDictsOnServerToSetDlg();
              checkDictsAvail();
              break;
          }
          break;
        case 420:                               // Server temporarly unavailable
          iDictQueryErr = iRCode;
          break;
        case 421:                               // Server shutting down at operator request
          iDictQueryErr = iRCode;
          break;
        case 530:                               // Access denied
          iDictQueryErr = iRCode;
          break;
        case 531:                               // Access denied, use "SHOW INFO" for server information
          iDictQueryErr = iRCode;
          break;
        case 532:                               // Access denied, unknown mechanism
          iDictQueryErr = iRCode;
          break;
        case 550:                               // Invalid database, use "SHOW DB" for list of databases
        //iD = sRead.indexOf( " " );
          break;
        case 551:                               // Invalid strategy, use "SHOW STRAT" for list of strategies
          //iD = sRead.indexOf( " " );
          break;
        case 552:                               // No match
          lookUpWordLoopNext();
          break;
        case 554:                               // No database present
          checkDictsAvail();
          //iD = sRead.indexOf( " " );
          break;
        case 555:                               // No strategies available
          //iD = sRead.indexOf( " " );
          break;
      }
    } else {                                    // lines that have not " " at 4th position = content or list entry
      switch( iResType ) {
        case 1:                                     // 1=databases - answer to "SHOW DB"
        dictsEntry.selected = false;
        iD = sRead.indexOf( " " );
        sD = sRead.left( iD );
        if( sD.left( 1 ) != "." ) {
          dictsEntry.name = sRead.left( iD );             // get database name
          sRead.remove( 0, iD );
          dictsEntry.desc = sRead.trimmed();              // get database description
          dictsOnServer.append( dictsEntry );
        }
        break;
        case 2:                                 // 2=database infos - answer to "SHOW INFO"
          break;
        case 3:                                 // 3=word match orig - original word for translation to follow acc dictd
          sRead = sRead.trimmed();
          lookup.orig = sRead.trimmed();
          ssRead = ssRead + "lookup.orig= " + sRead.trimmed() + "\n";
          iResType = 4;
          break;
        case 4:                                 // 4=word match trans - translation for this original word acc dictd
          if( sRead.isEmpty() == false ) {
            if( sRead.left(1) != "." ) {              // not end of one result entry/match => ignore line with only '.'
              sRead = sRead.trimmed();
              lookup.trans = lookup.trans + sRead;
            } else {
              lookups.append( lookup );
              iResType = 0;                     // needed for last line?????????????????????????
//          ssRead = ssRead + sRead + "\n";
            }
          }
          break;
      }
    }
  }
}

/*
// ****************************************************************************
// **  errors                                                                **
// ****************************************************************************
void dict::emitConErr()                     // emit connection errors, if dict server reachable but access denied ***************??????????????????????  needed ??
{
  sDictd = sRead;                           // need to save sRead immidiately where emitConErr() is called? **************************
  iConState = 0;
  emit sigDicErrQuery( iDictQueryErr, sDictd );
}
*/

// ********************************************************************
// **  check if all requested dictionaries are available             **
// ********************************************************************
void dict::checkDictsAvail()                // check if any of the selected dictionaries are available, send status update to qvt
{
//  cout << "checkDictsAvail(): reached !" << endl;                                     // *****  debug  *****
  dictsToQuery.clear();
  sErr = "";
  bSomeDictsUnAvail = false;
  if( dictsInConf.count() > 0 ) {
    if( dictsOnServer.count() > 0 ) {
      for ( i = 0; i < dictsInConf.count(); i++ ) {                     // list with dictionaries requested to query from config file
        bDictHaveThis = false;
        for ( i2 = 0; i2 < dictsOnServer.count(); i2++ ) {              // list with dictionaries available on the dict server
          if( dictsInConf.at( i ).name == dictsOnServer.at( i2 ).name ) {
            bDictHaveThis = true;
            dictsEntry.selected = true;
            dictsEntry.name = dictsInConf.at( i ).name;
            dictsEntry.desc = dictsOnServer.at( i2 ).desc;
            dictsToQuery.append( dictsEntry );                          // dict server that can be used, match between dictsConf and dictsAvail
          }
        }
        if( bDictHaveThis == false ) {                                  // this dictionary is no longer available!
          bSomeDictsUnAvail = true;
          sErr = sErr + dictsInConf.at( i ).name + "\n";
        }
      }

      if( bSomeDictsUnAvail == true ) {
        sErr = tr( "The following dictionary(ies) are no longer available on the dict server:\n \n" ) + sErr;
      }

      iDS = 0;                                  // set dict server state, 0=disabled, 1=no dict or none of selected, 2=some selected avail, 3= all selected avail
      if( dictsToQuery.count() > 0 ) {
        if( bSomeDictsUnAvail == false ) {        // all dicts avail
          iDS = 3;
          if( iTask == 3 ) {
            emit sigDicSendDictsOnServer( dictsOnServer, dictsToQuery );
          }
        } else {
          iDS = 2;                                // some dicts avail
          sErr = tr( "Error !\nSome dictionaries requested in the settings dialog to be used for word look ups are no longer available: \n \n" ) + sErr;
        }
      } else {
        iDS = 1;                                  // error, no dicts avail
        sErr = tr( "Not one of the selected dictionaries is available on this dict server!\n"
                   "Dict look up will be disabled.\n"
                   "It can be re-enabled in the settings dialog.\n \n"
                   "In case you query your own dict server:\n"
                   "Did you already install any dictionary(ies) beside the dict server itself?\n"
                   "If not, exit and query your distribution package manager with a search criteria like: dict\n"
                   "and install the free dictionary(ies) you need.\n \n" ) + sErr;
      }
    } else {
      iDS = 0;
      sErr = tr( "There is not a single dictionary available on the dict server!\n"
                 "Dict look up will be disabled.\n"
                 "It can be re-enabled in the settings dialog.\n \n"
                 "In case you query your own dict server:\n"
                 "Did you already install any dictionary(ies) beside the dict server itself?\n"
                 "If not, exit and query your distribution package manager with a search criteria like: dict\n"
                 "and install the free dictionary(ies) you need.\n \n" );
    }
  } else {
    iDS = 0;
    sErr = tr( "There is not a single dictionary selected for use in the settings dialog!\n"
               "This must never happen. Anyway, Dict look up will be disabled.\n"
               "It can be re-enabled in the settings dialog." );
  }
//  cout << "checkDictsAvail(): iDS= " << QString::number( iDS ).toStdString() << endl;                                     // *****  debug  *****
  if( iTask != 3 ) {
    emit sigDicSendDictsStatusUpd( iDS, sErr );
  }
}


#include "dict.moc"
